# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import subprocess
import threading

from ovos_utils.log import LOG


class TemperatureMonitorThread(threading.Thread):
    def __init__(self, fan_obj=None):
        self.fan_obj = fan_obj or FanControl()
        self.exit_flag = threading.Event()
        self._max_fanless_temp = 60.0
        threading.Thread.__init__(self)

    def run(self):
        LOG.debug("temperature monitor thread started")
        while not self.exit_flag.wait(30):
            LOG.debug(f"CPU temperature is {self.fan_obj.get_cpu_temp()}")

            current_temp = self.fan_obj.get_cpu_temp()
            if current_temp < self._max_fanless_temp:
                # Below specified fanless temperature
                fan_speed = 0
                LOG.debug(f"Temp below {self._max_fanless_temp}")
            elif current_temp > 80.0:
                LOG.warning("Thermal Throttling")
                fan_speed = 100
            else:
                # Specify linear fan curve inside normal operating temp range
                speed_const = 100/(80.0-self._max_fanless_temp)
                fan_speed = speed_const * (current_temp -
                                           self._max_fanless_temp)
                LOG.info(f"temp={current_temp}")

            LOG.info(f"Setting fan speed to: {fan_speed}")
            self.fan_obj.set_fan_speed(fan_speed)


class FanControl:
    # hardware speed range is appx 30-255
    # we convert from 0 to 100
    HDW_MIN = 0
    HDW_MAX = 255
    SFW_MIN = 0
    SFW_MAX = 100

    def __init__(self):
        self.fan_speed = 0
        self.set_fan_speed(self.fan_speed)

    def execute_cmd(self, cmd):
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        out, err = process.communicate()

        try:
            out = out.decode('utf8')
        except:
            pass

        try:
            err = err.decode('utf8')
        except:
            pass

        return out, err

    def cToF(self, temp):
        return (temp * 1.8) + 32

    def speed_to_hdw_val(self, speed):
        out_steps = self.HDW_MAX - self.HDW_MIN
        in_steps = self.SFW_MAX - self.SFW_MIN
        ratio = out_steps / in_steps
        # force compliance
        if speed > self.SFW_MAX:
            speed = self.SFW_MAX
        if speed < self.SFW_MIN:
            speed = self.SFW_MIN

        return int((speed * ratio) + self.HDW_MIN)

    def hdw_val_to_speed(self, hdw_val):
        out_steps = self.SFW_MAX - self.SFW_MIN
        in_steps = self.HDW_MAX - self.HDW_MIN
        ratio = out_steps / in_steps
        # force compliance
        if hdw_val > self.HDW_MAX:
            hdw_val = self.HDW_MAX
        if hdw_val < self.HDW_MIN:
            hdw_val = self.HDW_MIN

        return int(round(((hdw_val - self.HDW_MIN) * ratio) + self.SFW_MIN, 0))

    def hdw_set_speed(self, hdw_speed):
        # force compliance
        if hdw_speed > self.HDW_MAX:
            hdw_speed = self.HDW_MAX
        if hdw_speed < self.HDW_MIN:
            hdw_speed = self.HDW_MIN

        hdw_speed = str(hdw_speed)
        cmd = ["i2cset", "-a", "-y", "1", "0x04", "101", hdw_speed, "i"]
        out, err = self.execute_cmd(cmd)

    def set_fan_speed(self, speed):
        self.fan_speed = self.speed_to_hdw_val(speed)
        self.hdw_set_speed(self.fan_speed)

    def get_fan_speed(self):
        return self.hdw_val_to_speed(self.fan_speed)

    def get_cpu_temp(self):
        cmd = ["cat", "/sys/class/thermal/thermal_zone0/temp"]
        out, err = self.execute_cmd(cmd)
        return float(out.strip()) / 1000


class MycroftFan:
    """ abstract base class for a Mycroft Fan
     all fan classes must provide at least
     these basic methods """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def set_fan_speed(self, new_speed):
        """takes in value between 0 - 100
           converts to internal format"""
        return

    @abc.abstractmethod
    def get_fan_speed(self):
        """returns value between 0-100"""

    @abc.abstractmethod
    def get_cpu_temp(self):
        """returns temp in celsius"""
        return



# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import RPi.GPIO as GPIO
import time
from ovos_utils.log import LOG


class MycroftSwitch:
    """ abstract base class for a Mycroft Switch Array
     all switches must provide at least these basic methods """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def terminate(self):
        """terminates any underlying threads"""
        return

    @abc.abstractmethod
    def handle_action(self):
        return

    @abc.abstractmethod
    def handle_voldown(self):
        return

    @abc.abstractmethod
    def handle_volup(self):
        return

    @abc.abstractmethod
    def handle_mute(self, val):
        return

    @abc.abstractmethod
    def get_capabilities(self):
        return


class Switch:
    """
    Class to handle the Mark2 switches.
    Note - a switch is an abstract concept
    which applies to buttons and switches.
    The Mark2 actually has 4 different switches.
    Three buttons (volume up, down and activate)
    and a mute mic switch. All are read only
    and interrupt driven. Also note switches are
    pulled up so the active state is actually zero.
    """
    # GPIO pin numbers
    """ old sj201 mappings
    _SW_ACTION = 22
    _SW_VOL_UP = 23
    _SW_VOL_DOWN = 24
    _SW_MUTE = 25
    """
    # sj201Rev4
    _SW_VOL_UP = 22
    _SW_VOL_DOWN = 23
    _SW_ACTION = 24
    _SW_MUTE = 25

    _XMOS_POWER = 16  # Enable1V
    _XMOS_RESET = 27  # Reset XMOS

    def __init__(self, debounce=100):
        self.debounce = debounce
        self.active = 0

        # some switch implementations require a thread
        # we don't but we must meet the base requirement
        self.thread_handle = None

        self.capabilities = {
            "user_volup_handler": "button",
            "user_voldown_handler": "button",
            "user_action_handler": "button",
            "user_mute_handler": "slider"
        }

        # use BCM GPIO pin numbering
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)

        """
        # xmos related
        GPIO.setup(self._XMOS_POWER, GPIO.OUT)
        GPIO.setup(self._XMOS_RESET, GPIO.OUT)

        # power up the xmos
        #self.reset_xmos()
        time.sleep(0.001)
        GPIO.output(self._XMOS_POWER, 1)
        time.sleep(0.001)
        GPIO.output(self._XMOS_RESET, 1)
        """

        # we need to pull up the 3 buttons
        GPIO.setup(self._SW_ACTION, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.setup(self._SW_VOL_UP, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.setup(self._SW_VOL_DOWN, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.setup(self._SW_MUTE, GPIO.IN, pull_up_down=GPIO.PUD_UP)

        # establish default values
        self.SW_ACTION = GPIO.input(self._SW_ACTION)
        self.SW_VOL_UP = GPIO.input(self._SW_VOL_UP)
        self.SW_VOL_DOWN = GPIO.input(self._SW_VOL_DOWN)
        self.SW_MUTE = GPIO.input(self._SW_MUTE)

        # establish default handlers for each switch
        self.action_handler = self.handle_action
        self.vol_up_handler = self.handle_vol_up
        self.vol_down_handler = self.handle_vol_down
        self.mute_handler = self.handle_mute

        # attach callbacks
        GPIO.add_event_detect(self._SW_ACTION,
                              GPIO.BOTH,
                              callback=self.action_handler,
                              bouncetime=debounce)

        GPIO.add_event_detect(self._SW_VOL_UP,
                              GPIO.BOTH,
                              callback=self.vol_up_handler,
                              bouncetime=debounce)

        GPIO.add_event_detect(self._SW_VOL_DOWN,
                              GPIO.BOTH,
                              callback=self.vol_down_handler,
                              bouncetime=debounce)

        GPIO.add_event_detect(self._SW_MUTE,
                              GPIO.BOTH,
                              callback=self.mute_handler,
                              bouncetime=debounce)

        # user overides
        self.user_voldown_handler = None
        self.user_volup_handler = None
        self.user_action_handler = None
        self.user_mute_handler = None

    def get_capabilities(self):
        return self.capabilities

    def handle_action(self, channel):
        self.SW_ACTION = GPIO.input(self._SW_ACTION)
        if self.SW_ACTION == self.active:
            if self.user_action_handler is not None:
                self.user_action_handler()

    def handle_vol_up(self, channel):
        self.SW_VOL_UP = GPIO.input(self._SW_VOL_UP)
        if self.SW_VOL_UP == self.active:
            if self.user_volup_handler is not None:
                self.user_volup_handler()

    def handle_vol_down(self, channel):
        self.SW_VOL_DOWN = GPIO.input(self._SW_VOL_DOWN)
        if self.SW_VOL_DOWN == self.active:
            if self.user_voldown_handler is not None:
                self.user_voldown_handler()

    def handle_mute(self, channel):
        # No idea why this delay is necessary, but it makes the muting reliable
        time.sleep(0.05)
        self.SW_MUTE = GPIO.input(self._SW_MUTE)

        if self.user_mute_handler is not None:
            self.user_mute_handler(self.SW_MUTE)

    # recycle xmos power
    def reset_xmos(self):
        LOG.info("switch_gpio: reset_xmos() hit")
        """
        GPIO.output(self._XMOS_RESET, 0)
        time.sleep(0.001)
        GPIO.output(self._XMOS_POWER, 0)
        time.sleep(0.001)
        GPIO.output(self._XMOS_POWER, 1)
        time.sleep(0.001)
        GPIO.output(self._XMOS_RESET, 1)
        """

    def terminate(self):
        LOG.info("switch_gpio: terminate hit, calling GPIO.cleanup()")
        # GPIO.cleanup()



# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import threading
import typing
from queue import Queue

import time
from ovos_utils.log import LOG
from smbus2.smbus2 import SMBus, I2C_SMBUS_BLOCK_MAX


class Palette:
    BLACK = (0, 0, 0)
    WHITE = (255, 255, 255)
    YELLOW = (255, 255, 0)
    RED = (255, 0, 0)
    GREEN = (0, 255, 0)
    BLUE = (0, 0, 255)
    MAGENTA = (255, 0, 255)
    BURNT_ORANGE = (173, 64, 0)
    MYCROFT_RED = (216, 17, 89)
    MYCROFT_GREEN = (64, 219, 176)
    MYCROFT_BLUE = (34, 167, 240)


class MycroftLed:
    """ abstract base class for a Mycroft Led
     all leds must provide at least these basic methods """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self):
        return

    @abc.abstractmethod
    def set_led(self, which_led, color, immediate):
        """Set the color of a specific LED.
           Arguments:
            which_led (Int): the index of the LED to be changed
            color (Tuple): the RGB color as a three integer Tuple
            immediate (Bool): whether to change now or wait for the next call of show()
        """
        return

    @abc.abstractmethod
    def fill(self, color):
        """set all leds to the supplied color
           Arguments:
            color (Tuple): the RGB color as a three integer Tuple
        """
        return

    @abc.abstractmethod
    def show(self):
        """update leds from buffered data"""
        return

    @abc.abstractmethod
    def get_led(self, which_led):
        """returns current buffered value"""
        return

    @abc.abstractmethod
    def set_leds(self, leds):
        """updates buffer from leds and update hardware
           Arguments:
            leds [(of tuples),()]: the RGB color as a three integer Tuple
        """
        return

    @abc.abstractmethod
    def get_capabilities(self):
        """returns capabilities object"""
        return


class Led(MycroftLed):
    real_num_leds = 12  # physical
    num_leds = 12  # logical
    black = (0, 0, 0)  # TODO pull from palette
    device_addr = 0x04

    def __init__(self):
        self.bus = SMBus(1)
        self.brightness = 0.5
        self.capabilities = {
            "num_leds": self.num_leds,
            "brightness": "(0.0-1.0)",
            "led_colors": "MycroftPalette",
            "reserved_leds": list(
                range(self.num_leds, self.real_num_leds)
            ),
        }

    def adjust_brightness(self, cval, bval):
        return min(255, cval * bval)

    def get_capabilities(self):
        return self.capabilities

    def _set_led(self, pixel, color):
        """ internal interface
            permits access to the
            reserved leds """
        red_val = int(color[0])
        green_val = int(color[1])
        blue_val = int(color[2])

        # cmd =   "i2cset -y 1 %d %d %d %d %d i " % (
        #             self.device_addr,
        #             pixel,
        #             red_val,
        #             green_val,
        #             blue_val)
        # os.system(cmd)
        # LOG.debug("Execute %s" % (cmd,))

        self.bus.write_i2c_block_data(
            self.device_addr, pixel,
            [red_val, green_val, blue_val]
        )

    def _set_led_with_brightness(self, pixel, color, blevel):
        self._set_led(
            pixel,
            list(
                map(
                    self.adjust_brightness,
                    color,
                    (blevel,) * 3)))

    def show(self):
        """ show buffered leds, only used
           for older slower devices """
        pass

    def set_led(self, pixel, color):
        """ external interface enforces led
            reservation and honors brightness """
        self._set_led(
            pixel % self.num_leds,
            list(
                map(
                    self.adjust_brightness,
                    color,
                    (self.brightness,) * 3)))

    def fill(self, color):
        """fill all leds with the same color"""

        rgb = [int(self.adjust_brightness(c, self.brightness))
               for c in color[:3]]
        try:
            # Each element in rgb is a 3 byte tuple
            led_per_block = int(str(I2C_SMBUS_BLOCK_MAX / 3).split('.')[0])
        except Exception as e:
            LOG.exception(e)
            led_per_block = 10
        leds_to_write = self.num_leds
        last_written_idx = 0
        LOG.debug(f"Writing {leds_to_write} LEDs in blocks of {led_per_block}")
        while leds_to_write > led_per_block:
            leds_to_write = leds_to_write - led_per_block
            self.bus.write_i2c_block_data(
                self.device_addr, last_written_idx,
                rgb * led_per_block
            )
            last_written_idx += led_per_block
        if leds_to_write > 0:
            self.bus.write_i2c_block_data(
                self.device_addr, last_written_idx,
                rgb * leds_to_write
            )

    def set_leds(self, new_leds):
        """set leds from tuple array"""
        for x in range(0, self.num_leds):
            self.set_led(x, new_leds[x])


class LedAnimation:
    """Base class for LED animations"""

    def __init__(self, led_obj):
        self.led_obj = led_obj

    def start(self):
        """Begin LED animation"""
        pass

    def step(self, context: typing.Dict[str, typing.Any]) -> bool:
        """Single step of the animation.

        Put time.sleep inside here.

        Arguments:
            context: dict with user-defined values

        Returns:
            True if animation should continue
        """
        return False

    def stop(self):
        """End LED animation"""
        pass


class PulseLedAnimation(LedAnimation):
    def __init__(self, led_obj, pal_obj):
        super().__init__(led_obj)

        self.pal_obj = pal_obj
        self.exit_flag = False
        self.color_tup = self.pal_obj.MYCROFT_GREEN
        self.delay = 0.1
        self.brightness = 100
        self.step_size = 5
        self.tmp_leds = []

    def start(self):
        self.led_obj.fill(self.color_tup)

        self.brightness = 100
        self.led_obj.brightness = self.brightness / 100
        self.led_obj.fill(self.color_tup)

    def step(self, context):
        if (self.brightness + self.step_size) > 100:
            self.brightness = self.brightness - self.step_size
            self.step_size = self.step_size * -1

        elif (self.brightness + self.step_size) < 0:
            self.brightness = self.brightness - self.step_size
            self.step_size = self.step_size * -1

        else:
            self.brightness += self.step_size

        self.led_obj.brightness = self.brightness / 100
        self.led_obj.fill(self.color_tup)

        time.sleep(self.delay)
        return True

    def stop(self):
        self.led_obj.brightness = 1.0
        self.led_obj.fill(self.pal_obj.BLACK)


class ChaseLedAnimation(LedAnimation):
    def __init__(self, led_obj, background_color, foreground_color):
        super().__init__(led_obj)

        self.bkgnd_col = background_color
        self.fgnd_col = foreground_color
        self.exit_flag = False
        self.color_tup = foreground_color
        self.delay = 0.1

    def start(self):
        self.led_obj.fill(self.fgnd_col)

    def step(self, context):
        fgnd_col = context.get("chase.foreground_color", self.fgnd_col)
        bkgnd_col = context.get("chase.background_color", self.bkgnd_col)
        stop = context.get("chase.stop", False)

        for x in range(0, 10):
            self.led_obj.set_led(x, fgnd_col)
            time.sleep(self.delay)
            self.led_obj.set_led(x, bkgnd_col)

        if stop:
            return False

        return True

    def stop(self):
        self.led_obj.fill(self.led_obj.black)


class LedThread(threading.Thread):
    def __init__(self, led_obj, animations=None):
        self.led_obj = led_obj
        self.animations = animations or {}
        self.queue = Queue()
        self.animation_running = False
        self.animation_name = None
        self._context: {}

        super().__init__()

    def start_animation(self, name: str):
        self.stop_animation()
        self.queue.put(name)

    def stop_animation(self, name=None):
        if name and (self.animation_name != name):
            # Different animation is playing
            return

        self.animation_running = False

    @property
    def context(self):
        return self._context

    def run(self):
        try:
            while True:
                self.animation_name = None
                self.animation_running = False

                name = self.queue.get()
                current_animation = self.animations.get(name)

                if current_animation is not None:
                    try:
                        self._context = {}
                        self.animation_name = name
                        self.animation_running = True
                        current_animation.start()
                        while self.animation_running and current_animation.step(
                                context=self._context
                        ):
                            time.sleep(0)
                        current_animation.stop()
                    except Exception:
                        self.led_obj.fill(self.led_obj.black)
                        LOG.exception("error running animation '%s'", name)

                else:
                    LOG.error("No animation named %s", name)
        except Exception:
            LOG.exception("error running led animation")



from ovos_PHAL.detection import is_mycroft_sj201
from ovos_PHAL_plugin_mk2.fan import FanControl, TemperatureMonitorThread
from ovos_PHAL_plugin_mk2.leds import ChaseLedAnimation, LedAnimation, \
    PulseLedAnimation, Led, Palette, LedThread
from ovos_PHAL_plugin_mk2.switch import Switch
from ovos_plugin_manager.phal import PHALPlugin
from mycroft_bus_client.message import Message
import time
from ovos_utils.log import LOG


class MycroftMark2Validator:
    @staticmethod
    def validate(config=None):
        # check i2c to determine if sj201 is connected
        return is_mycroft_sj201()


class MycroftMark2(PHALPlugin):
    validator = MycroftMark2Validator

    def __init__(self, bus=None, config=None):
        super().__init__(bus=bus, name="ovos-PHAL-plugin-mk2", config=config)
        self.fan = FanControl()
        self.leds = Led()
        self.switches = Switch()
        self.switches.user_mute_handler = self._on_mute
        self._last_mute = -1
        self.switches.user_action_handler = self.on_button_press
        self.switches.user_voldown_handler = self.on_button_voldown_press
        self.switches.user_volup_handler = self.on_button_volup_press
        self._last_press = 0

        # start the temperature monitor thread
        self.temperatureMonitorThread = TemperatureMonitorThread(self.fan)
        self.temperatureMonitorThread.start()

        # init leds all turned off
        self.turn_off_leds()
        self.led_thread = LedThread(self.leds,
                                    animations={
                                        "pulse": PulseLedAnimation(self.leds, Palette),
                                        "chase": ChaseLedAnimation(self.leds,
                                                                   background_color=Palette.BLUE,
                                                                   foreground_color=Palette.BLACK,
                                                                   ),
                                    })
        self.led_thread.start()
        if self.switches.SW_MUTE == 1:
            self.on_hardware_mute()

    def shutdown(self):
        self.temperatureMonitorThread.exit_flag.set()
        super().shutdown()

    def _on_mute(self, val):
        LOG.debug("Mark2:HardwareEnclosure:handle_mute() - val = %s" % (val,))
        if val != self._last_mute:
            self._last_mute = val
            if val == 0:
                self.on_hardware_unmute()
            else:
                self.on_hardware_mute()

    def on_button_press(self):
        LOG.debug("SJ201 Listen button pressed")
        # debounce this 10 seconds
        if time.time() - self._last_press > 10:
            self._last_press = time.time()
            self.bus.emit(Message("mycroft.mic.listen"))

    def on_button_volup_press(self):
        LOG.debug("SJ201 VolumeUp button pressed")
        self.bus.emit(Message("mycroft.volume.increase"))

    def on_button_voldown_press(self):
        LOG.debug("SJ201 VolumeDown button pressed")
        self.bus.emit(Message("mycroft.volume.decrease"))

    def on_hardware_mute(self):
        """Called when hardware switch is set to mute"""
        # Triggers red border
        self.bus.emit(Message("mycroft.mic.mute"))
        self.leds.fill(Palette.BURNT_ORANGE)

    def on_hardware_unmute(self):
        """Called when hardware switch is set to unmute"""
        # Removes red border
        self.bus.emit(Message("mycroft.mic.unmute"))
        self.turn_off_leds()

    def turn_off_leds(self):
        self.leds.fill(Palette.BLACK)

    # Audio Events
    def on_awake(self, message=None):
        ''' on wakeup animation '''
        # TODO new led animation

    def on_sleep(self, message=None):
        ''' on naptime animation '''
        # TODO new led animation
        self.turn_off_leds()

    def on_reset(self, message=None):
        """The enclosure should restore itself to a started state.
        Typically this would be represented by the eyes being 'open'
        and the mouth reset to its default (smile or blank).
        """
        self.turn_off_leds()

    def on_record_begin(self, message=None):
        self.led_thread.start_animation("pulse")

    def on_record_end(self, message=None):
        self.led_thread.stop_animation("pulse")

    # System Events
    def on_system_reset(self, message=None):
        """The enclosure hardware should reset any CPUs, etc."""
        self.turn_off_leds()
